/*  console.c -- Console interface for Intercom
    Copyright (C) 2001-2003  Shane Wegner
    This file is part of Intercom.

    Intercom is free software; you can redistribute it and/or modify it
    under the terms of version 2 of the GNU General Public License as
    published by the Free Software Foundation.

    Intercom is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of version 2 of the GNU General Public
    License along with Intercom; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    To contact the author, please send email to shane@cm.nu.
    */

/* $Id: console.c,v 1.65 2003/02/13 20:22:49 shane Exp $ */

    /* AIX requires this to be the first thing in the file. */
#ifdef _AIX
  #pragma alloca
#endif

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <time.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "subst/getopt.h"
#endif
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#if defined(HAVE_SYS_SOUNDCARD_H)
#include <sys/soundcard.h>
#elif defined(HAVE_MACHINE_SOUNDCARD_H)
#include <machine/soundcard.h>
#else
#include <soundcard.h>
#endif
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#ifdef HAVE_LIBREADLINE
#ifdef HAVE_READLINE_READLINE_H
#include <readline/readline.h>
#else
#include <readline.h>
#endif
#endif
#ifdef HAVE_LIBHISTORY
#ifdef HAVE_READLINE_HISTORY_H
#include <readline/history.h>
#else
#include <history.h>
#endif
#endif
#ifdef USE_CRYPTO
#include <openssl/sha.h>
#endif

#include "intercom.h"
#include "protocol.h"
#include "console.h"
#include "userdefs.h"
#include "hooks.h"
#include "iwordexp.h"
#ifdef USE_CRYPTO
#include "crypto.h"
#endif
#ifdef USE_GETTEXT
#include <libintl.h>
#define _(String) gettext(String)
#define gettext_noop(String) (String)
#define N_(String) gettext_noop(String)
#else
#define _(String) (String)
#define N_(String) (String)
#endif

struct console_cmd {
const char *cmd;
char *desc;
void (*func)(int, char **);
};

struct alias {
char *cmd;
char *newcmd;
struct alias *next;
};

static struct alias *alias_list = NULL;

#ifndef HAVE_LIBREADLINE
static int console_active = 0;
#endif
#ifdef HAVE_LIBHISTORY
char *historyfile = NULL;
#endif

static void cmd_alias(int ARGC, char **ARGV)
{
static const struct option long_options[]={
{"help", no_argument, NULL, 'h'},
{"unalias", no_argument, NULL, 'u'},
{"verbose", no_argument, NULL, 'v'},
{NULL, 0, NULL, 0}};
int o, verbose = 0, unalias = 0, all = 0;
struct alias *ap, *next, *prev = NULL;
optind = 0;
while ((o = getopt_long(ARGC, ARGV, "ahuv", long_options, (int *)0)) != EOF) {
switch(o) {
case 'a':
all = 1;
break;
case 'h':
printf(_("Usage: %s [options] [<alias>] {<command>]\n\n"
"Valid options:\n"
"-a, --all\tWhen used with -u, removes all aliases"
"-h, --help\tDisplay this help information\n"
"-u, --unalias\tRemove an alias\n"
"-v, --verbose\tPrint alias with value when creating an alias\n\n"
"If <command> is omited, the alias is printed with it\'s command.\n"
"If <alias> is omited, all aliases and their values will be printed.\n\n"),
ARGV[0]);
return;
case 'u':
unalias = 1;
break;
case 'v':
verbose = 1;
break;
default:
fprintf(stderr, _("Try '%s --help' for a usage summary.\n"), ARGV[0]);
return;
}
}
if (unalias) {
if (!all && (ARGC - optind) == 0) {
fprintf(stderr, _("%s: Nothing to unalias.\n"), ARGV[0]);
return;
} else if ((ARGC - optind) > 1) {
fprintf(stderr, _("%s: Too many command-line options specified.\n"), ARGV[0]);
return;
}
for (ap = alias_list; ap != NULL; ap = next) {
next = ap->next;
if (all || !strcmp(ap->cmd, ARGV[optind])) {
if (prev == NULL)
alias_list = ap->next;
else
prev->next = ap->next;
if (verbose)
printf(_("Alias '%s' removed.\n"), ap->cmd);
free(ap->cmd);
free(ap->newcmd);
free(ap);
if (!all)
break;
} else
prev = ap;
}
return;
} /* endif unalias */
if (!(ARGC - optind)) {
fputs(_("Currently defined aliases:\n"), stdout);
for (ap = alias_list; ap != NULL; ap = ap->next)
printf("'%s' = '%s'\n", ap->cmd, ap->newcmd);
putchar('\n');
} else if ((ARGC - optind) == 1) {
for (ap = alias_list; ap != NULL; ap = ap->next)
if (!strcmp(ap->cmd, ARGV[optind])) {
printf(_("'%s' aliases '%s'\n"), ap->cmd, ap->newcmd);
break;
}
} else {
for (ap = alias_list; ap != NULL; ap = ap->next) {
if (!strcmp(ap->cmd, ARGV[optind])) {
if (verbose)
fprintf(stderr, _("Alias '%s' already defined.\n"), ap->cmd);
return;
}
}
if (alias_list != NULL) {
for (ap = alias_list; ap->next != NULL; ap = ap->next)
;
ap->next = malloc(sizeof(struct alias));
ap = ap->next;
} else {
ap = malloc(sizeof(struct alias));
alias_list = ap;
}
ap->cmd = strdup(ARGV[optind]);
ap->newcmd = strdup(ARGV[optind+1]);
ap->next = NULL;
if (verbose)
printf(_("'%s' aliases '%s'\n"), ap->cmd, ap->newcmd);
}
}

void cmd_answer(int ARGC, char **ARGV)
{
const struct option long_options[] = {
{"compression", required_argument, NULL, 'c'},
{"rate", required_argument, NULL, 'r'},
{"help", no_argument, NULL, 'h'},
{NULL, 0, NULL, 0}};
int o, i;
struct pkt_callresponse callpkt;
struct pkt_audiostart audiopkt;
struct sockaddr_in r_address;
callpkt.aparams.compression = call.rcv_aparams.compression;
callpkt.aparams.rate = call.rcv_aparams.rate;
callpkt.aparams.bitwidth = call.rcv_aparams.bitwidth;
callpkt.responsecode = CALLRESPONSE_ACCEPT;
optind = 0;
while ((o = getopt_long(ARGC, ARGV, "c:hr:", long_options, (int *)0)) != EOF) {
switch(o) {
case 'c':
if ((i = int_compression(optarg)) == -1) {
fprintf(stderr, _("%s: Compression format %s is invalid or unsupported.\n"),
ARGV[0], optarg);
return;
} else
callpkt.aparams.compression = i;
break;
case 'h':
printf(_("Usage: %s [options]\n\n"
"Valid options:\n"
"-c, --compression\tSet audio compression format\n"
"-r, --rate\t\tSet audio sample rate\n"
"-h, --help\t\tDisplay this help information\n"
"\n"), ARGV[0]);
return;
case 'r':
callpkt.aparams.rate = atoi(optarg);
break;
default:
fprintf(stderr, _("Try '%s --help' for a usage summary.\n"), ARGV[0]);
return;
}
}
if ((ARGC - optind)) {
fprintf(stderr, _("%s: Too many arguments on command-line.\n"), ARGV[0]);
return;
} else if (!TEST_BIT(call.state, CALL_STATE_WFRANSWER)) {
fprintf(stderr, _("%s: No incoming call is pending at this time.\n"), ARGV[0]);
return;
}
memcpy(&call.snd_aparams, &callpkt.aparams, sizeof(callpkt.aparams));
#ifdef USE_CRYPTO
if (call.cryptinfo.type != CALL_CRYPT_NONE) {
char *pass;
char clr_challenge[128];
char challenge_sha1[20];
size_t try_nr, l;
pass = opt_get_str("crypt_passphrase");
if (*pass) {
cr_set_key(pass);
cr_decrypt(call.cryptinfo.challenge, 128, clr_challenge, &l);
SHA1(clr_challenge, 128, challenge_sha1);
if (!memcmp(challenge_sha1, call.cryptinfo.challenge_sha1, 20)) {
puts(_("Default passphrase is good."));
goto goodpass;
}
}
for (try_nr = 1; try_nr <= 3; try_nr++) {
pass = getpass(_("Enter passphrase for encrypted call: "));
cr_set_key(pass);
cr_decrypt(call.cryptinfo.challenge, 128, clr_challenge, &l);
SHA1(clr_challenge, 128, challenge_sha1);
if (!memcmp(challenge_sha1, call.cryptinfo.challenge_sha1, 20))
goto goodpass;
printf(_("%s: Incorrect passphrase (try %d)\n"),
ARGV[0], try_nr);
}
fprintf(stderr, _("%s: Closing connection.\n"), ARGV[0]);
closecall();
}
goodpass:
#endif
fputs(_("Answering call.\n"), stdout);
queuepacket(&call.cssq, PKT_TYPE_CALLRESPONSE, (char *)&callpkt, sizeof(callpkt));
hk_trigger(hk_call_connect);
CLEAR_BIT(call.state, CALL_STATE_WFRANSWER);
SET_BIT(call.state, CALL_STATE_INCALL);
call.connstats.start_time = time(NULL);
if ((call.dsock = setup_udpsock(&r_address)) == -1) {
closecall();
return;
}
if (*opt_get_str("snd_play_device")) {
if ((call.pdsp = dsp_setup(opt_get_str("snd_play_device"),
(opt_get_int("snd_singleopen")) ? O_RDWR : O_WRONLY, &call.rcv_aparams)) == -1)
return;
if (ioctl(call.pdsp, SNDCTL_DSP_GETBLKSIZE, &call.pdsp_blksize) == -1) {
perror("ioctl(SND_DSP_GETBLKSIZE)");
return;
}
call.pdsp_buff = malloc(call.pdsp_blksize);
iob_setminbytes(&call.pdspq, call.pdsp_blksize * 4);
}
audiopkt.dport = r_address.sin_port;
queuepacket(&call.cssq, PKT_TYPE_AUDIOSTART, (char *)&audiopkt, sizeof(audiopkt));
SET_BIT(call.state, CALL_STATE_AUDIORCV);
}

void cmd_call(int ARGC, char **ARGV)
{
const struct option long_options[] = {
{"compression", required_argument, NULL, 'c'},
#ifdef USE_CRYPTO
{"encrypt", optional_argument, NULL, 'e'},
#endif
{"port", required_argument, NULL, 'p'},
{"rate", required_argument, NULL, 'r'},
{"user", required_argument, NULL, 'u'},
{"help", no_argument, NULL, 'h'},
{NULL, 0, NULL, 0}};
int o, i;
struct pkt_call callpkt;
struct pkt_hello hellopkt;
struct sockaddr_in r_address;
int port = opt_get_int("port");
callpkt.cryptinfo.type = CALL_CRYPT_NONE;
callpkt.aparams.compression = int_compression(opt_get_str("snd_compression"));
callpkt.aparams.rate = opt_get_int("snd_rate");
callpkt.aparams.bitwidth = opt_get_int("snd_bitwidth");
hellopkt.proto_version = INTERCOM_PROTOCOL_VERSION;
strcpy(hellopkt.username, opt_get_str("username"));
strcpy(hellopkt.clientname, "Intercom " VERSION);
optind = 0;
while ((o = getopt_long(ARGC, ARGV, "c:e::hp:r:u:", long_options, (int *)0)) != EOF) {
switch(o) {
case 'c':
if ((i = int_compression(optarg)) == -1) {
fprintf(stderr, _("%s: Compression format %s is invalid or unsupported.\n"),
ARGV[0], optarg);
return;
} else
callpkt.aparams.compression = i;
break;
case 'e': {
#ifndef USE_CRYPTO
fprintf(stderr,
_("%s: error: -e switch used but crypto has been disabled in this build\n"), ARGV[0]);
return;
#else
int t;
if (optarg == NULL)
t = cr_type_fromstr(opt_get_str("crypt_method"));
else
t = cr_type_fromstr(optarg);
if (t == -1) {
fprintf(stderr, _("%s: Invalid or unsupported encryption method.\n"),
ARGV[0]);
return;
} else
callpkt.cryptinfo.type = t;
#endif
}
break;
case 'h':
printf(_("Usage: %s [options] host\n\n"
"Valid options:\n"
"-c, --compression\tSet audio compression format\n"
"-e, --encrypt\tEncrypt audio data optionally a crypt method can be specified\n"
"-p, --port\t\tConnect on specified port\n"
"-r, --rate\t\tSet audio sample rate\n"
"-u, --user\t\tSpecifies the username to send\n"
"-h, --help\t\tDisplay this help information\n"
"\n"), ARGV[0]);
return;
case 'p':
port = atoi(optarg);
break;
case 'r':
callpkt.aparams.rate = atoi(optarg);
break;
case 'u':
strncpy(hellopkt.username, optarg, 128);
break;
default:
fprintf(stderr, _("Try '%s --help' for a usage summary.\n"), ARGV[0]);
return;
}
}
if ((ARGC - optind) != 1) {
fprintf(stderr, _("%s: No hostname specified.\n"), ARGV[0]);
return;
} else if ((ARGC - optind) > 1) {
fprintf(stderr, _("%s: Too many options on command-line.\n"), ARGV[0]);
return;
} else if (TEST_BIT(call.state, CALL_STATE_CONNECTED)) {
fprintf(stderr, _("%s: Another call is already active.\n"), ARGV[0]);
return;
}
printf(_("%s: Placing call to '%s'.\n"), ARGV[0], ARGV[optind]);
if ((call.csock = do_connect(ARGV[optind], port, &r_address)) == -1)
return;
memcpy(&call.r_address, &r_address, sizeof(r_address));
icsetenv("remote_hostip", inet_ntoa(r_address.sin_addr));
if (opt_get_int("reverse_lookup")) {
if ((call.r_hostent = gethostbyaddr((char *)&r_address.sin_addr, sizeof(r_address.sin_addr), AF_INET)) == NULL)
call.r_hostent = gethostbyname(inet_ntoa(r_address.sin_addr));
icsetenv("remote_hostname", call.r_hostent->h_name);
} else
call.r_hostent = gethostbyname(inet_ntoa(r_address.sin_addr));
memcpy(&call.snd_aparams, &callpkt.aparams, sizeof(callpkt.aparams));
queuepacket(&call.cssq, PKT_TYPE_HELLO, (char *)&hellopkt, sizeof(hellopkt));
#ifdef USE_CRYPTO
if (callpkt.cryptinfo.type != CALL_CRYPT_NONE) {
char *pass, *pass2;
char challenge[128];
int i;
size_t os;
pass2 = opt_get_str("crypt_passphrase");
if (!*pass2) {
pass = getpass(_("Enter encryption passphrase: "));
pass = strdup(pass);
pass2 = getpass(_("Confirm passphrase: "));
if (strcmp(pass, pass2)) {
fprintf(stderr, _("%s: passphrases do not match, aborting.\n"), ARGV[0]);
free(pass);
close(call.csock);
call.csock = -1;
return;
}
free(pass);
}
for (i = 0; i < 128; i++)
challenge[i] = (int)(256.0*rand()/(RAND_MAX+1.0));
SHA1(challenge, 128, callpkt.cryptinfo.challenge_sha1);
call.cryptinfo.type = callpkt.cryptinfo.type;
cr_set_key(pass2);
cr_encrypt(challenge, 128, callpkt.cryptinfo.challenge, &os);
}
memcpy(&call.cryptinfo, &callpkt.cryptinfo, sizeof(callpkt.cryptinfo));
#endif
queuepacket(&call.cssq, PKT_TYPE_CALL, (char *)&callpkt, sizeof(callpkt));
SET_BIT(call.state, CALL_STATE_CONNECTED|CALL_STATE_CALLING);
}

static void cmd_copyright(int ARGC, char **ARGV)
{
fputs("Intercom - Real-Time audio communication utility.\n\n"
"Copyright (C) 2001-2003 Shane Wegner\n\n"
"\n"
"This program is free software; you can redistribute it and/or\n"
"modify it under the terms of version 2 of the GNU General Public\n"
"License as published by the Free Software Foundation.\n"
"\n"
"This program is distributed in the hope that it will be useful,\n"
"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
"GNU General Public License for more details.\n"
"\n"
"A copy of version 2 of the GNU General Public License is\n"
"distributed with Intercom under the filename COPYING.\n"
"\n"
"For license details use the command: \"copyright\"\n\n", stdout);
}

static void cmd_exec(int ARGC, char **ARGV)
{
if (ARGC != 2)
fprintf(stderr, _("Usage: %s <command>\n"), ARGV[0]);
else
system(ARGV[1]);
}

static void cmd_hangup(int ARGC, char **ARGV)
{
if (ARGC > 1)
fprintf(stderr, _("Usage: '%s'.\n"), ARGV[0]);
else if (!TEST_BIT(call.state, CALL_STATE_CONNECTED))
fprintf(stderr, _("%s: No call is currently in progress.\n"), ARGV[0]);
else {
puts(_("Hanging up."));
if (TEST_BIT(call.state, CALL_STATE_WFRANSWER)) {
/* A call was requested and rejected so send back a response packet */
/* before hanging up. */
if (fcntl(call.csock, F_SETFL, O_RDWR) == -1)
perror("fcntl");
else {
struct pkt_callresponse pkt;
size_t buffsize;
char *buff;
pkt.responsecode = CALLRESPONSE_DENY;
queuepacket(&call.cssq, PKT_TYPE_CALLRESPONSE, (char *)&pkt, sizeof(pkt));
buffsize = iob_buffsize(&call.cssq);
buff = alloca(buffsize);
iob_read(&call.cssq, buff, buffsize);
write(call.csock, buff, buffsize);
} /* endif fcntl */
} /* endif rejected */
if (TEST_BIT(call.state, CALL_STATE_INCALL))
hk_trigger(hk_hangup_lc);
closecall();
}
}

/* cmd_help is defined below */
static void cmd_help(int, char **);

static void cmd_hook(int ARGC, char **ARGV)
{
struct option long_options[] = {
{"help", no_argument, NULL, 'h'},
{"verbose", no_argument, NULL, 'v'},
{NULL, 0, NULL, 0}};
int o;
int verbose = 0;
optind = 0;
while ((o = getopt_long(ARGC, ARGV, "hv", long_options, (int *)0)) != EOF)
switch(o) {
case 'h':
printf(_("Valid usage: %s [options]\n"
"%s [options] add <type> <cmd>\n"
"%s clearid <id>\n"
"%s cleartype <type>\n"
"%s clearall\n"
"\nOptions:\n"
"-h, --help\tDisplay this help information\n"
"-v, --verbose\tVerbose output\n"
"\n"), ARGV[0], ARGV[0], ARGV[0], ARGV[0], ARGV[0]);
return;
case 'v':
verbose = 1;
break;
default:
fprintf(stderr, _("%s: Try '%s --help' for a usage summary.\n"),
ARGV[0], ARGV[0]);
return;
};
if ((ARGC - optind) == 0) { /* list */
struct hook_info *h;
puts(_("Currently defined event-hooks:"));
for (h = hook_list; h != NULL; h = h->next) {
printf("Id: %d, Type: ", h->id);
switch(h->type) {
case hk_call_incoming:
fputs("call_incoming", stdout);
break;
case hk_call_outgoing:
fputs("call_outgoing", stdout);
break;
case hk_call_connect:
fputs("call_connect", stdout);
break;
case hk_hangup_lc:
fputs("hangup_lc", stdout);
break;
case hk_hangup_rmt:
fputs("hangup_rmt", stdout);
break;
case hk_sighup:
fputs("sighup", stdout);
break;
default:
assert(0);
};
printf(", Cmd: '%s'\n", h->cmd);
}
putchar('\n');
}
else if (!strcasecmp(ARGV[optind], "add")) {
hk_type_t type;
if ((ARGC - optind) != 3) {
fprintf(stderr, _("%s: Invalid usage.  Try '%s --help' for a usage summary.\n"),
ARGV[0], ARGV[0]);
return;
}
if ((type = hk_strtotype(ARGV[optind+1])) == hk_invalid) {
fprintf(stderr, _("%s: Invalid event-hook type.\n"), ARGV[0]);
return;
}
hk_add(type, ARGV[optind + 2]);
} else if (!strcasecmp(ARGV[optind], "clearall")) {
if ((ARGC - optind) != 1) {
fprintf(stderr, _("%s: Invalid usage.  Try '%s --help' for a usage summary.\n"),
ARGV[0], ARGV[0]);
return;
}
o = hk_clearall();
if (verbose)
printf(_("%s: Deleted %d hooks\n"), ARGV[0], o);
} else if (!strcasecmp(ARGV[optind], "clearid")) {
int id;
if ((ARGC - optind) != 2) {
fprintf(stderr, _("%s: Invalid usage.  Try '%s --help' for a usage summary.\n"),
ARGV[0], ARGV[0]);
return;
}
id = atoi(ARGV[optind + 1]);
o = hk_clear(id);
if (verbose)
printf(_("%s: Deleted %d hooks\n"), ARGV[0], o);
} else if (!strcasecmp(ARGV[optind], "cleartype")) {
hk_type_t type;
if ((ARGC - optind) != 2) {
fprintf(stderr, _("%s: Invalid usage.  Try '%s --help' for a usage summary.\n"),
ARGV[0], ARGV[0]);
return;
}
if ((type = hk_strtotype(ARGV[optind + 1])) == hk_invalid) {
fprintf(stderr, "%s: Invalid event-hook type.\n", ARGV[0]);
return;
}
o = hk_cleartype(type);
if (verbose)
printf(_("%s: Deleted %d hooks\n"), ARGV[0], o);
} else
fprintf(stderr,
_("%s: Invalid hook command.  Try '%s --help' for a usage summary.\n"),
ARGV[0], ARGV[0]);
}

static void cmd_if(int ARGC, char **ARGV)
{
struct option long_options[]={
{"help", no_argument, NULL, 'h'},
{"verbose", no_argument, NULL, 'v'},
{NULL, 0, NULL, 0}};
int o;
int verbose = 0;
int my_code = 0;
const char *op = "=";
const char *my_cmd, *my_action, *my_elaction = NULL;
int do_action = 0;
optind = 0;
while ((o = getopt_long(ARGC, ARGV, "c:hv", long_options, (int *)0)) != EOF)
switch(o) {
case 'h':
printf(_("Usage: %s [options] <command> [<op> <exitcode> | && | ||] <action> else <action>\n\n"
"Options\n"
"-h, --help\tDisplay this help information\n"
"-v, --verbose\tTurn on verbose output\n\n"
"Valid values for <op> are =, !=, <, >, <=, and >=.\n\n"),ARGV[0]);
return;
case 'v':
verbose = 1;
break;
default:
fprintf(stderr, _("Try '%s --help' for a usage summary.\n"), ARGV[0]);
return;
};
if ((ARGC - optind) < 2) {
fprintf(stderr, _("%s: Too few arguments.  Try '%s --help' for a usage summary."),
ARGV[0], ARGV[0]);
return;
}
my_cmd = ARGV[optind++];
if (!strcmp(ARGV[optind], "=") || !strcmp(ARGV[optind], "!=") ||
!strcmp(ARGV[optind], "<") || !strcmp(ARGV[optind], ">") ||
!strcmp(ARGV[optind], "<=") || !strcmp(ARGV[optind], ">=")) {
op = ARGV[optind++];
if (!(ARGC - optind)) {
fprintf(stderr, _("%s: Parse error near '%s'\n"),
ARGV[0], ARGV[optind-1]);
return;
}
my_code = atoi(ARGV[optind++]);
} else if (!strcmp(ARGV[optind], "&&"))
optind++; /* default */
else if (!strcmp(ARGV[optind], "||")) {
op = "!=";
optind++;
}
if (!(ARGC - optind)) {
fprintf(stderr, _("%s: Parse error near '%s'\n"),
ARGV[0], ARGV[optind-1]);
return;
}
my_action = ARGV[optind++];
if ((ARGC - optind)) {
if (strcasecmp(ARGV[optind++], "else")) {
fprintf(stderr, _("%s: Parse error near '%s'.\n"), ARGV[0], ARGV[optind-1]);
return;
}
if (!(ARGC - optind)) {
fprintf(stderr, _("%s: Parse error near '%s'\n"),
ARGV[0], ARGV[optind-1]);
return;
}
my_elaction = ARGV[optind++];
}
if ((o = system(my_cmd)) == -1) {
fprintf(stderr, _("%s: system() failed: %s\n"),
ARGV[0], strerror(errno));
return;
}
o = WEXITSTATUS(o);
if (o == 127) {
fprintf(stderr, _("%s: System command failed.\n"), ARGV[0]);
return;
}
if (verbose)
printf(_("%s: Command exited with status %d comparing to %d with %s\n"),
ARGV[0], o, my_code, op);
if (!strcasecmp(op, "="))
do_action = (o == my_code);
if (!strcasecmp(op, "!="))
do_action = (o != my_code);
if (!strcasecmp(op, "<"))
do_action = (o < my_code);
if (!strcasecmp(op, ">"))
do_action = (o > my_code);
if (!strcasecmp(op, "<="))
do_action = (o <= my_code);
if (!strcasecmp(op, ">="))
do_action = (o >= my_code);
if (do_action) {
if (verbose)
printf(_("%s: Running '%s'\n"), ARGV[0], my_action);
processline(0, my_action);
} else if (my_elaction) {
if (verbose)
printf(_("%s: Running '%s'\n"), ARGV[0], my_elaction);
processline(0, my_elaction);
}
}

static void cmd_message(int ARGC, char **ARGV)
{
char *msgdata, *msgp;
size_t i, msglen = 0;
if (ARGC == 1) {
fprintf(stderr, _("%s: No message to send.\n"), ARGV[0]);
return;
}
if (!TEST_BIT(call.state, CALL_STATE_CONNECTED)) {
fprintf(stderr, _("%s: There is no connection currently established.\n"), ARGV[0]);
return;
}
for (i = 1; i < ARGC; i++)
msglen += strlen(ARGV[i]) + 1;
msgdata = alloca(msglen);
msgp = msgdata;
for (i = 1; i < ARGC; i++)
msgp += sprintf(msgp, "%s ", ARGV[i]);
queuepacket(&call.cssq, PKT_TYPE_USRTEXT, msgdata, msglen-1);
puts(_("Your message has been queued."));
}

static void cmd_quit(int ARGC, char ** ARGV)
{
if (ARGC > 1)
fprintf(stderr, _("Usage: '%s'.\n"), ARGV[0]);
else
done = 1;
console_done();
}

static void cmd_read(int ARGC, char **ARGV)
{
static const struct option long_options[]={
{"help", no_argument, NULL, 'h'},
{"silent", no_argument, NULL, 's'},
{NULL, 0, NULL, 0}};
int silent = 0;
size_t i, l;
char buff[4096];
char *startp;
FILE *f;
optind = 0;
while ((i = getopt_long(ARGC, ARGV, "hs", long_options, (int *)0)) != EOF) {
switch(i) {
case 'h':
printf(_("Usage: %s [options] <filename> [<filename> ...]\n\n"
"Valid options:\n"
"-h, --help\tOutput this help information\n"
"-s, --silent\tDo not output error or status messages\n"
"\n"), ARGV[0]);
return;
case 's':
silent = 1;
break;
default:
fprintf(stderr, _("Try '%s --help' for a usage summary.\n"), ARGV[0]);
return;
}
}
if (!(ARGC - optind)) {
if (!silent)
fprintf(stderr, _("%s: No file(s) were specified.\n"), ARGV[0]);
return;
}
for (i = optind; i < ARGC; i++) {
if ((f = fopen(ARGV[i], "r")) == NULL) {
if (!silent)
fprintf(stderr, _("fopen(%s): %s\n"), ARGV[i], strerror(errno));
continue;
}
while(fgets(buff, 4096, f) != NULL) {
l = strlen(buff) - 1;
while(buff[l] == '\r' || buff[l] == '\n')
buff[l--] = 0;
for (startp = buff; *startp == ' ' || *startp == '\t'; startp++)
;
if (*startp != '#')
processline(0, startp);
}
fclose(f);
}
}

static void escapestr(const char *from, char *to)
{
const char *fp;
char *tp = to;
for (fp = from; *fp; fp++) {
if (*fp == '"' || *fp == '\\') {
*tp = '\\';
tp++;
}
*tp = *fp;
tp++;
}
*tp = 0;
}

static void cmd_save(int ARGC, char **ARGV)
{
static const struct option long_options[]={
{"aliases", no_argument, NULL, 'a'},
{"file", required_argument, NULL, 'f'},
{"help", no_argument, NULL, 'h'},
{"hooks", no_argument, NULL, 'H'},
{"variables", no_argument, NULL, 'v'},
{NULL, 0, NULL, 0}};
char *filename = NULL, *key = NULL, *val = NULL;
FILE *f;
int o;
int novariable = 0, noalias = 0, nohooks = 0;
struct alias *ap;
struct hook_info *h;
time_t ti;
struct tm *tm;
optind = 0;
while ((o = getopt_long(ARGC, ARGV, "af:hv", long_options, (int *)0)) != EOF) {
switch(o) {
case 'a':
novariable = 1;
nohooks = 1;
break;
case 'f':
filename = optarg;
break;
case 'h':
printf(_("Usage: %s [options]\n\n"
"Valid options:\n"
"-a, --aliases\tOnly save aliases to rc-file\n"
"-f, --file\tSpecify where configuration data is to be written\n"
"-h, --help\tDisplay this help information\n"
"-H, --hooks\tOnly save event-hooks\n"
"-v, --variables\tONly save configuration variables to rc-file\n\n"),
ARGV[0]);
return;
case 'H':
noalias = 1;
novariable = 1;
break;
break;
case 'v':
noalias = 1;
nohooks = 1;
break;
default:
fprintf(stderr, _("Try '%s --help' for a usage summary.\n"), ARGV[0]);
return;
}
}
if ((ARGC - optind) > 0) {
fprintf(stderr, _("%s: Too many options specified.  Try '%s --help' for a usage summary.\n"),
ARGV[0], ARGV[0]);
return;
}
if (filename == NULL) {
char *home = getenv("HOME");
if (home == NULL) {
fputs(_("HOME environment variable not set and no -f option specified.\n"), stderr);
return;
}
filename = alloca(strlen(home) + 13);
sprintf(filename, "%s/.intercomrc", home);
}
if ((f = fopen(filename, "w")) == NULL) {
fprintf(stderr, _("fopen(%s): %s\n"), filename, strerror(errno));
return;
}
val = malloc(64);
ti = time(NULL);
tm = localtime(&ti);
strftime(val, 64, "%Y-%m-%d %H:%M:%S", tm);
fprintf(f, "# This file was generated by intercom version " VERSION "\n"
"# on %s.\n\n",
val);
if (!novariable) {
fputs("# Configuration variables\n", f);
for (o = 0; useropts[o].key != NULL; o++) {
if (!strcmp(useropts[o].def_value, useropts[o].value))
continue;
key = realloc(key, strlen(useropts[o].key) * 2 + 1);
val = realloc(val, strlen(useropts[o].value) * 2 + 1);
escapestr(useropts[o].key, key);
escapestr(useropts[o].value, val);
if (useropts[o].desc != NULL)
fprintf(f, "# %s\n", _(useropts[o].desc));
fprintf(f, "set \"%s\" \"%s\"\n", key, val);
}
fputs("\n", f);
}
if (!noalias) {
fputs("# Command aliases\n", f);
for (ap = alias_list; ap != NULL; ap = ap->next) {
key = realloc(key, strlen(ap->cmd) * 2 + 1);
val = realloc(val, strlen(ap->newcmd) * 2 + 1);
escapestr(ap->cmd, key);
escapestr(ap->newcmd, val);
fprintf(f, "alias \"%s\" \"%s\"\n", key, val);
}
fputs("\n", f);
}
if (!nohooks) {
fputs("# Event-hooks\n", f);
fputs("hook clearall\n", f);
for (h = hook_list; h != NULL; h = h->next) {
fputs("hook add ", f);
switch(h->type) {
case hk_call_incoming:
fputs("call_incoming", f);
break;
case hk_call_outgoing:
fputs("call_outgoing", f);
break;
case hk_call_connect:
fputs("call_connect", f);
break;
case hk_hangup_lc:
fputs("hangup_lc", f);
break;
case hk_hangup_rmt:
fputs("hangup_rmt", f);
break;
case hk_sighup:
fputs("sighup", f);
break;
default:
assert(0);
};
val = realloc(val, strlen(h->cmd) * 2 + 1);
escapestr(h->cmd, val);
fprintf(f, " \"%s\"\n", val);
}
fputs("\n", f);
}
if (key != NULL)
free(key);
if (val != NULL)
free(val);
fclose(f);
printf(_("Configuration settings written to '%s'\n"), filename);
}

static void cmd_set(int ARGC, char **ARGV)
{
static const struct option long_options[]={
{"help", no_argument, NULL, 'h'},
{"verbose", no_argument, NULL, 'v'},
{NULL, 0, NULL, 0}};
int o;
int verbose = 0;
optind = 0;
while ((o = getopt_long(ARGC, ARGV, "hv", long_options, (int *)0)) != EOF) {
switch(o) {
case 'h':
printf(_("Usage: %s [options] [<variable>] [<value>]\n\n"
"Valid options:\n"
"-h, --help\tDisplay this help information\n"
"-v, --verbose\tDisplay a variable\'s new value when set\n\n"
"If <value> is omited, the current value is displayed.\n"
"If <variable> is omited, all variables with their values are displayed.\n\n"),
ARGV[0]);
return;
case 'v':
verbose = 1;
break;
default:
fprintf(stderr, _("Try '%s --help' for a usage summary.\n"), ARGV[0]);
return;
}
}
if (!(ARGC - optind)) {
puts(_("Currently defined variables:\n"));
for (o = 0; useropts[o].key != NULL; o++)
switch(useropts[o].type) {
case opt_int:
case opt_float:
printf("%s = %s (%s)\n", useropts[o].key, useropts[o].value, useropts[o].desc);
break;
case opt_str:
printf("%s = '%s' (%s)\n", useropts[o].key,
(!strcmp(useropts[o].key, "crypt_passphrase")) ? _("(not shown)") :
useropts[o].value, useropts[o].desc);
break;
default:
assert(0);
}
putchar('\n');
} else if ((ARGC - optind) == 1) {
o = opt_find(ARGV[optind]);
if (o == -1)
fprintf(stderr, _("%s: No such variable (%s).\n"), ARGV[0], ARGV[optind]);
else
switch(useropts[o].type) {
case opt_int:
case opt_float:
printf("%s = %s\n", useropts[o].key, useropts[o].value);
break;
case opt_str:
printf("%s = '%s'\n", useropts[o].key,
(!strcmp(useropts[o].key, "crypt_passphrase")) ? _("(not shown)") :
useropts[o].value);
break;
default:
assert(0);
}
} else if ((ARGC - optind) == 2) {
o = opt_find(ARGV[optind]);
if (o == -1)
fprintf(stderr, _("%s: No such variable (%s).\n"), ARGV[0], ARGV[optind]);
else
switch(useropts[o].type) {
case opt_int:
opt_set_int(ARGV[optind], atoi(ARGV[optind+1]));
break;
case opt_float:
opt_set_float(ARGV[optind], strtod(ARGV[optind+1], NULL));
break;
case opt_str:
opt_set_str(ARGV[optind], ARGV[optind+1]);
break;
default:
assert(0);
}
if (strcasecmp(ARGV[optind], "crypt_passphrase"))
icsetenv(ARGV[optind], ARGV[optind+1]);
} else
fprintf(stderr, _("%s: Too many command-line options. Try '%s --help' for a usage summary."),
ARGV[0], ARGV[0]);
}

static char *munge(int v)
{
int m = 0;
int div = 1;
/* TODO: fix this obvious memory leak.  I tried *
 * Making it a static char [64] but that always returned the same value. *
 * I think gcc might be optimizing where it shouldn't be or some such thing. */
char *s = malloc(16);
if (v >= 1073741824) {
div = 1073741824;
m = 'g';
} else if (v >= 1048576) {
div = 1048576;
m = 'm';
} else if (v >= 1024) {
div = 1024;
m = 'k';
}
if (m)
sprintf(s, "%0.2f%c", (float)v / div, m);
else
sprintf(s, "%d", v);
return s;
}

static void cmd_stats(int ARGC, char **ARGV)
{
audio_buf_info info;
time_t t;
struct tm *tm;
char *s;
if (ARGC > 1) {
fprintf(stderr, _("Usage: '%s'.\n"), ARGV[0]);
return;
}
if (!TEST_BIT(call.state, CALL_STATE_CONNECTED)) {
fprintf(stderr, _("%s: There is no call currently in progress.\n"), ARGV[0]);
return;
}
t = time(NULL) - call.connstats.start_time;
tm = gmtime(&t);
printf(_("Current connection statistics:\n"
"Call duration: %02d:%02d:%02d\n"
"Control packets sent: %u (%u bytes)\n"
"Control packets received: %u (%u bytes)\n\n"
"Audio packets sent: %u (%sb, %sb/sec)\n"
"Audio packets received: %u (%sb, %sb/sec)\n"
"Audio packets lost: %u (%u out of sequence)\n\n"),
tm->tm_hour, tm->tm_min, tm->tm_sec,
call.connstats.cpkts_sent, call.connstats.cbytes_sent,
call.connstats.cpkts_recv, call.connstats.cbytes_recv,
call.connstats.apkts_sent, munge(call.connstats.abytes_sent),
munge(call.connstats.abytes_sent / t),
call.connstats.apkts_recv, munge(call.connstats.abytes_recv),
munge(call.connstats.abytes_recv / t),
call.connstats.apkts_lost, call.connstats.apkts_order);
if (TEST_BIT(call.state, CALL_STATE_AUDIOSND)) {
s = str_compression(call.snd_aparams.compression);
printf(_("Sending audio at %u KHZ, %u bits with compression codec %s.\n"),
call.snd_aparams.rate, call.snd_aparams.bitwidth, s);
if (ioctl(call.rdsp, SNDCTL_DSP_GETISPACE, &info) == -1)
perror("SNDCTL_DSP_GETISPACE");
else
printf(_("Fragment size: %d, Number of fragments: %d\n"),
info.fragsize, info.fragstotal);
}
if (TEST_BIT(call.state, CALL_STATE_AUDIORCV)) {
s = str_compression(call.rcv_aparams.compression);
printf(_("Receiving audio at %u KHZ, %u bits with compression codec %s.\n"),
call.rcv_aparams.rate, call.rcv_aparams.bitwidth, s);
if (ioctl(call.pdsp, SNDCTL_DSP_GETOSPACE, &info) == -1)
perror("SNDCTL_DSP_GETOSPACE");
else
printf(_("Fragment size: %d, Number of fragments: %d, bytes in output buffer: %d\n"),
info.fragsize, info.fragstotal, iob_buffsize(&call.pdspq));
}
putchar('\n');
}

static const struct console_cmd cmd_list[]={
{"alias", N_("Manage command aliases."), cmd_alias},
{"answer", N_("Answer an incoming call."), cmd_answer},
{"call", N_("Place a call to a remote host."), cmd_call},
{"copyright", N_("Show copyright and license details."), cmd_copyright},
{"exec", N_("Execute external program."), cmd_exec},
{"hangup", N_("Terminate the current call."), cmd_hangup},
{"help", N_("Gives basic help information for valid commands."), cmd_help},
{"hook", N_("Administer event-hooks"), cmd_hook},
{"if", N_("Execute intercom command based on exit-code of shell command"), cmd_if},
{"message", N_("Send a text message to connected party."), cmd_message},
{"quit", N_("Exit from Intercom."), cmd_quit},
{"read", N_("Read commands from a file."), cmd_read},
{"save", N_("Save configuration information."), cmd_save},
{"set", N_("Show or set configuration variables."), cmd_set},
{"stats", N_("Show call statistics."), cmd_stats},
{NULL, NULL, NULL},
};

static void cmd_help(int ARGC, char **ARGV)
{
size_t i, l;
static size_t maxlen = 0;
if (ARGC > 1) {
char line[strlen(ARGV[1]) + 8];
sprintf(line, "%s --help", ARGV[1]);
processline(0, line);
} else {
/* First pass to get maxlen value */
if (!maxlen) {
for (i = 0; cmd_list[i].cmd != NULL; i++) {
l = strlen(cmd_list[i].cmd);
if (l > maxlen)
maxlen = l;
}
}
puts(_("Valid commands are:"));
for (i = 0; cmd_list[i].cmd != NULL; i++) {
printf("%-*s %s\n", maxlen, cmd_list[i].cmd, _(cmd_list[i].desc));
}
putchar('\n');
}
}

void processline(int addhistory, const char *line)
{
wordexp_t wv;
int r;
static int depth = 0;
size_t i;
struct alias *ap;
if (depth > opt_get_int("cmd_maxdepth")) {
fputs(_("Error, command or alias loop detected.\n"), stderr);
return;
}
depth++;
#if (DEBUG >= 1)
if (verbose)
printf(_("debug: processline: processing line '%s' at depth %d.\n"), line, depth);
#endif
#ifdef HAVE_LIBHISTORY
if (addhistory && *line)
add_history(line);
#endif
if ((r = iwordexp(line, &wv, WRDE_UNDEF))) {
switch(r) {
case WRDE_BADCHAR:
fputs(_("Bad character in input.\n"), stderr);
break;
case WRDE_BADVAL:
fputs(_("Undefined shell variable in input.\n"), stderr);
break;
case WRDE_CMDSUB:
fputs(_("Command-substitution is not enabled.\n"), stderr);
break;
case WRDE_NOSPACE:
fputs(_("wordexp: Memory allocation failure.\n"), stderr);
break;
case WRDE_SYNTAX:
fputs(_("Input syntax error.\n"), stderr);
break;
default:
fputs(_("wordexp: Input error: unknown error.\n"), stderr);
} /* switch */
} else if (wv.we_wordc) {
#if (DEBUG >= 2)
if (verbose) {
for (i = 0; i < wv.we_wordc; i++)
printf("Debug: wordex word %d = '%s'\n", i, wv.we_wordv[i]);
}
#endif
for (ap = alias_list; ap != NULL; ap = ap->next) {
if (!strcmp(wv.we_wordv[0], ap->cmd)) {
size_t len = strlen(ap->newcmd) + 1, pos = 0, i2;
char *newline;
for (i2 = 1; i2 < wv.we_wordc; i2++)
len += strlen(wv.we_wordv[i2]) + 1;
newline = malloc(len);
pos += sprintf(newline, "%s", ap->newcmd);
for (i2 = 1; i2 < wv.we_wordc; i2++)
pos += sprintf(&newline[pos], " %s", wv.we_wordv[i2]);
processline(0, newline);
goto out;
}
}
for (i = 0; cmd_list[i].cmd != NULL; i++)
if (!strcmp(wv.we_wordv[0], cmd_list[i].cmd)) {
cmd_list[i].func(wv.we_wordc, wv.we_wordv);
goto out;
}
fprintf(stderr,
_("Unknown command: '%s', try 'help' for a list of valid commands.\n"), wv.we_wordv[0]);
out:
iwordfree(&wv);
} else {
iwordfree(&wv);
} /* endif wordexp */
depth--;
}

static void rl_processline(char *line)
{
if (line == NULL) {
putchar('\n');
return;
}
processline(1, line);
free(line);
}

#ifndef HAVE_LIBREADLINE
char *readln(void)
{
char *line = malloc(1024);
size_t l;
fgets(line, 1024, stdin);
l = strlen(line);
if (!l) {
free(line);
return NULL;
}
if (line[l - 1] == '\n')
line[l - 1] = 0;
return line;
}
#endif

void console_processkey(void)
{
#ifdef HAVE_LIBREADLINE
rl_callback_read_char();
#else
/* If we're not using readline, the tty should be line buffered. */
rl_processline(readln());
if (console_active) {
printf("%s", opt_get_str("prompt"));
fflush(stdout);
}
#endif
}

void console_init(void)
{
#ifdef HAVE_LIBREADLINE
rl_callback_handler_install(opt_get_str("prompt"), rl_processline);
#ifdef HAVE_LIBHISTORY
using_history();
asprintf(&historyfile, "%s/.intercom.history", getenv("HOME"));
read_history(historyfile);
history_set_pos(history_length);
#endif
#else
console_active = 1;
printf("%s", opt_get_str("prompt"));
fflush(stdout);
#endif
}

void console_done(void)
{
#ifdef HAVE_LIBREADLINE
rl_callback_handler_remove();
#ifdef HAVE_LIBHISTORY
stifle_history(opt_get_int("history_limit"));
write_history(historyfile);
#endif
#else
console_active = 0;
#endif
}
